OLDC.HUD = {
	scroll = 320*FRACUNIT,
	timer = 0
}

local O = OLDC
local CV = O.CVars
local F = O.Functions

local hub = O.hub
local firstmap = O.firstmap
local defaultmap = O.defaultmap
local didmodifiedprint

local accept
local refuse
local bigaccept
local bigrefuse

hud.add(function(v)
	accept = $ or v.cachePatch("ACCEPT")
	refuse = $ or v.cachePatch("REFUSE")
	bigaccept = $ or v.cachePatch("BIGACCEPT")
	bigrefuse = $ or v.cachePatch("BIGREFUSE")
end)

local function R_WorldToScreen2(p, cam, target)

	-- local sx = cam.angle - R_PointToAngle2(p.mo.x, p.mo.y, target.x, target.y)
	local sx = cam.angle - R_PointToAngle(target.x, target.y)
	local visible = false

	-- Get the h distance from the target
	local hdist = R_PointToDist(target.x, target.y)
	-- print(AngleFixed(sx)/FU )
	if sx > ANGLE_90 or sx < ANGLE_270 then
		-- sx = 0 -- return {x=0, y=0, scale=0}
		visible = false
	else
		sx = FixedMul(160*FU, tan($1)) + 160*FU
		visible = true
	end

	-- local sx = 160*FU + (160 * tan(cam.angle - R_PointToAngle(target.x, target.y)))
	-- local sy = 100*FU + (100 * (tan(cam.aiming) - FixedDiv(target.z, hdist)))
	local sy = 100*FU + 160 * (tan(cam.aiming) - FixedDiv(target.z-cam.z, 1 + FixedMul(hdist, cos(cam.angle - R_PointToAngle(target.x, target.y))) ))
	
	-- local c = cos(p.viewrollangle)
	-- local s = sin(p.viewrollangle)
	-- sx = $1+FixedMul(c, target.x) + FixedMul(s, target.y)
	-- sy = $1+FixedMul(c, target.y) - FixedMul(s, target.x)

	local ss = FixedDiv(160*FU, hdist)

	return {x=sx, y=sy, scale=ss, inview=visible}
end

local function map(x, in_min, in_max, out_min, out_max)
	return out_min + (x - in_min)*(out_max - out_min)/(in_max - in_min)
end

local portals = {}

addHook("MapChange", function(map)
	portals = {}
end)

addHook("GameQuit", function()
	portals = nil
end)

addHook("NetVars", function(net)
	portals = net(portals)
end)

addHook("MobjSpawn", function(mo)
	if titlemapinaction
		return
	end
	table.insert(portals, mo)
end, MT_LEVELSPOT)

local function HUB_PortalTitleDrawer(v, stplyr, cam)
	-- TODO: could be better
	for i = #portals, 1, -1
		local mobj = portals[i]
		
		if not mobj.valid
			table.remove(portals, i)
			continue
		end
		
		local screencoords = R_WorldToScreen2(stplyr, camera or cam, mobj)

		if not screencoords.inview continue end
		
		-- Mobj angle
		local angle = (mobj.spawnpoint.angle)

		local oldcmap = mapheaderinfo[G_GetMapNumber("S0")+angle]
		
		-- Get together title strings
		-- (tostring fallback in the event someone tries to add maps over them and wastes the entire script in one go...)
		local title = tostring(oldcmap.lvlttl) or "???"
		local act = oldcmap.actnum
		local zone = (oldcmap.levelflags & LF_NOZONE) and "" or "Zone"

		-- Compiled title
		local portaltitle = string.format("%s %s", title, zone)

		portaltitle = $ and string.gsub($, "[^\32-\126]", "") .. (act and string.format(", Act %d", act) or "") or "\x82 STRING ERROR \x80"
		
		-- Get distance
		local dist = R_PointToDist2(stplyr.mo.x, stplyr.mo.y, mobj.x, mobj.y)
		local maxdist = 1024*FU

		if (dist < maxdist and P_CheckSight(stplyr.mo, mobj)) then
			local tr = {0, V_20TRANS, V_40TRANS, V_60TRANS, V_80TRANS}
			v.drawString(screencoords.x, (screencoords.y)-(16*FU), portaltitle, V_ALLOWLOWERCASE|tr[map(dist, 64*FU, maxdist, 1, 5)], "thin-fixed-center")
		end
	end
end

local function DrawRequested(v)
	local HUD = O.HUD
	
	local x = 8*FRACUNIT
	local y = 30*FRACUNIT
	local s = FRACUNIT
	local scroll = HUD.scroll
	local req = O.Data.voteInfo.player
	local f = V_SNAPTOTOP|V_SNAPTOLEFT
	
	local icon = v.getSprite2Patch(req.skin, SPR2_LIFE)
	OLDC_DrawFontString(v, x - abs(scroll), y, s/2, s/2, "Requested by", "RFNT", false, 4, f, 1, SKINCOLOR_WHITE, true)
	v.drawScaled(x - abs(scroll) + 8*FRACUNIT, y + 16*FRACUNIT, s, icon, f, v.getColormap(req.skin or "sonic", req.skincolor or SKINCOLOR_GREEN))
	OLDC_DrawFontString(v, x - abs(scroll) + 16*FRACUNIT, y + 8*FRACUNIT, s, s, req.name or "Brak Eggman", "RFNT", false, 8, f, 1, SKINCOLOR_YELLOW, true)
end

local function ScrambleText(v, str, n)
	local substr = ""
	local l = str:len()
	
	if n < l
		local start = l - n
		substr = str:sub(1, start)
		str = str:sub(start, l)
	end
	
	for i = 1, str:len()
		local c = str:sub(i, i)
		
		if c:byte() >= 48 and c:byte() <= 57
			substr = $ .. string.char(v.RandomRange(48, 57))
		elseif c:byte() >= 65 and c:byte() <= 90
			substr = $ .. string.char(v.RandomRange(65, 90))
		elseif c:byte() >= 97 and c:byte() <= 122
			substr = $ .. string.char(v.RandomRange(97, 122))
		else
			substr = $ .. c
		end
	end
	
	return substr
end

-- Converts numbers into the cute continue countdown patches
local function DigitToPatch(v, n)
	local d = n / 10
	local i = n - d * 10

	local patches = d and {
		v.cachePatch("COLRNUM"..d),
		v.cachePatch("COLRNUM"..i)
	} or {
		v.cachePatch("COLRNUM"..i)
	}

	return patches
end

local function DrawContinueNumber(v, x, y, s, c, a)
	local patches = DigitToPatch(v, c)
	local color = v.getColormap(TC_DEFAULT, SKINCOLOR_WHITE)
	local width = patches[1].width + (patches[2] and patches[2].width or 0)
	local cx = x
	
	if (a & V_PARAMMASK)
		cx = $ - width*s/(3 - (a & V_PARAMMASK))
	end

	for i, j in ipairs(patches)
		v.drawScaled(cx, y, s, j, a & ~V_PARAMMASK, v.getColormap(TC_DEFAULT, SKINCOLOR_WHITE))
		
		cx = $ + j.width*s
	end
end

local function GetMapName(map)
	local V = O.Data.voteInfo
	
	if not map
		if V.map
			map = V.map
		else
			map = defaultMap
		end
	end
		
	local name = mapheaderinfo[map].lvlttl
		
	if not (mapheaderinfo[map].levelflags & LF_NOZONE)
		name = $.." Zone"
	end
		
	if mapheaderinfo[map].actnum
		name = $..", Act "..mapheaderinfo[map].actnum
	end
		
	return name
end

local function GetVotePatch(p)
	if OLDC.Data.playerVotes[#p] == true
		return "ACCEPT"
	elseif OLDC.Data.playerVotes[#p] == false
		return "REFUSE"
	else
		return "DECIDE"
	end
end

local function DrawVoting(v, p)
	local HUD = O.HUD
	local D = O.Data
	local V = D.voteInfo
	
	if not V return end
	
	local t = HUD.timer
	local e = V.endtimer
	local i = HUD.inittimer
	local scroll = HUD.scroll
	
	-- Draw black fade
	v.fadeScreen(31, min(5, t/2))
	
	-- White fade for multiplayer
	if multiplayer and F.getVoteCount("yes") <= F.getVoteCount("no")
		v.fadeScreen(0, min(10, max(e, not V.voting and t or 0)))
	end
	
	local color = v.cachePatch("~138")
	local x = 160*FRACUNIT
	local y = 100*FRACUNIT
	local s = FixedDiv(min(TICRATE/3, t), TICRATE/3)
	local wscale = v.width()*s/64
	local hscale = 16*s/64
	
	-- Draw teal bar
	v.drawStretched(scroll < FRACUNIT and scroll * -1 or 0, y - 90*FRACUNIT + 8*FRACUNIT - 8*s, wscale, hscale, color, V_SNAPTOLEFT|V_SNAPTOTOP)
	
	local author = mapheaderinfo[V.map].authorname or "STJr"
	local zone = GetMapName() or "Gay Baby Jail Zone"
	local scrambletimer = string.len(zone) - max(0, (leveltime - i)/2 - TICRATE/8)
	local text = ScrambleText(v, zone, scrambletimer)
	
	-- Name of the level
	OLDC_DrawFontString(v, x + 15*scroll/9, y - 90*FRACUNIT, FRACUNIT/2, FRACUNIT/2,
		text, "LTFNT", true, 8, V_SNAPTOTOP, 0, SKINCOLOR_WHITE, true)
	-- Name of the entrant
	OLDC_DrawFontString(v, x + 15*scroll/9, y - 78*FRACUNIT, FRACUNIT, FRACUNIT,
		author, "RFNT", false, 4, V_SNAPTOTOP, 0, SKINCOLOR_YELLOW, true)
	
	-- Draw name of player who requested the vote
	if multiplayer then DrawRequested(v) end
	
	-- Draw clear sign
	v.drawScaled(320*FRACUNIT + abs(scroll), y - 74*FRACUNIT, FRACUNIT, v.cachePatch("CLEARBAR"), V_SNAPTOTOP|V_SNAPTORIGHT)
	v.drawScaled(320*FRACUNIT + abs(scroll) - 16*FRACUNIT, y - 64*FRACUNIT, FRACUNIT/2, v.cachePatch(D.mapChecks[V.map] and "MAPCLEAR" or "MAPNOCLEAR"), V_SNAPTOTOP|V_SNAPTORIGHT)
	
	local portal = v.cachePatch("RBYPRT"..((leveltime%20) + 1))
	local zigzag = v.cachePatch("LTACTBLU")
	local levelpic = (
		mapheaderinfo[V.map].staticoverlay
		and vid.cachePatch("MSTATIC"..((leveltime % 4)+1))
		or v.getSpritePatch(SPR_LVLA, V.map - firstmap)
	)
	
	local yes = D.playerVotes[#p] == true
	local no = D.playerVotes[#p] == false
	local sinparam = sin(FixedAngle(leveltime*FRACUNIT*8)) * 2
	local cosparam = cos(FixedAngle(leveltime*FRACUNIT*8)) * 2
	local yesx = x - 92*s
	local nox = x + 92*s
	local w
	
	-- Static circles
	v.drawScaled(yesx, y + 6*s, s/2, v.cachePatch("CIRCLE"), V_40TRANS, v.getColormap(TC_DEFAULT, SKINCOLOR_EMERALD))
	v.drawScaled(nox, y + 6*s, s/2, v.cachePatch("CIRCLE"), V_40TRANS, v.getColormap(TC_DEFAULT, SKINCOLOR_RED))
	
	-- Spinning circles
	if not yes or (yes and not (p.voteblink and leveltime & 2))
		v.drawScaled(
			yesx + sinparam * (yes and 2 or 1),
			y + 6*s + cosparam * (yes and 2 or 1),
			4*s/10,
			v.cachePatch("CIRCLE"),
			not yes and V_70TRANS or 0,
			v.getColormap(TC_DEFAULT, SKINCOLOR_MINT)
		)
	end
	
	if not no or (no and not (p.voteblink and leveltime & 2))
		v.drawScaled(
			nox + cosparam * (no and 2 or 1),
			y + 6*s + sinparam * (no and 2 or 1),
			4*s/10,
			v.cachePatch("CIRCLE"),
			not no and V_70TRANS or 0,
			v.getColormap(TC_DEFAULT, SKINCOLOR_SALMON)
		)
	end
	
	-- Icons
	v.drawScaled(yesx, y + 6*s, s, bigaccept)
	v.drawScaled(nox, y + 6*s, s, bigrefuse)
	
	-- Labels and button prompts
	w = OLDC_DrawFontString(v, yesx, y + 24*s, 3*s/2, 3*s/2, "Let's go!", "RFNT", false, 4, 0, 0, SKINCOLOR_WHITE, true)
	v.drawScaled(yesx - w/2 - 12*FRACUNIT, y + 32*s, s, v.cachePatch(leveltime/3 & 2 and "BUTTON1" or "BUTTON2"))
	w = OLDC_DrawFontString(v, nox, y + 24*s, 3*s/2, 3*s/2, "No way!", "RFNT", false, 4, 0, 0, SKINCOLOR_WHITE, true)
	v.drawScaled(nox + w/2 + 12*FRACUNIT, y + 32*s, s, v.cachePatch(leveltime/3 & 2 and "BUTTON3" or "BUTTON4"))
	
	-- Draw the portal graphic, and the level orb
	v.drawScaled(x, y, s, portal, 0, v.getColormap(TC_DEFAULT, SKINCOLOR_TEAL))
	v.drawScaled(x, y + 2*FRACUNIT, s, levelpic, V_20TRANS)
	
	-- Vote count
	if multiplayer
		local trans = (9 - FixedInt(8 * (sin(FixedAngle(leveltime*FRACUNIT*10))/2 + FRACUNIT/2))) * V_10TRANS
		local patch = v.cachePatch("SIDEBORDER")
		local color
		
		if F.getVoteCount("yes") > F.getVoteCount("no")
			color = v.getColormap(TC_DEFAULT, SKINCOLOR_EMERALD)
			v.drawScaled(0 - abs(scroll), 180*FRACUNIT, FRACUNIT, patch, trans|V_SNAPTOLEFT|V_SNAPTOBOTTOM, color)
		elseif F.getVoteCount("yes") < F.getVoteCount("no")
			color = v.getColormap(TC_DEFAULT, SKINCOLOR_RED)
			v.drawScaled(320*FRACUNIT + abs(scroll), 180*FRACUNIT, FRACUNIT, patch, trans|V_FLIP|V_SNAPTORIGHT|V_SNAPTOBOTTOM, color)
		end
		
		v.drawScaled(14*FRACUNIT - abs(scroll), 180*FRACUNIT, FRACUNIT, accept, V_SNAPTOLEFT|V_SNAPTOBOTTOM)
		DrawContinueNumber(v, 25*FRACUNIT - abs(scroll), 175*FRACUNIT, FRACUNIT/2, F.getVoteCount("yes"), V_SNAPTOLEFT|V_SNAPTOBOTTOM)
		
		v.drawScaled(306*FRACUNIT + abs(scroll), 180*FRACUNIT, FRACUNIT, refuse, V_SNAPTORIGHT|V_SNAPTOBOTTOM)
		DrawContinueNumber(v, 295*FRACUNIT + abs(scroll), 175*FRACUNIT, FRACUNIT/2, F.getVoteCount("no"), 2|V_SNAPTORIGHT|V_SNAPTOBOTTOM)
	end
	
	local maxtimer = (CV.vote.value > 0) and CV.vote.value * TICRATE or 12
	local wscale = v.width() * FixedDiv(V.timer, maxtimer)/64
	local y = min(t, 8)
	local dupy = y/2 * v.dupy()
	
	-- Timer
	v.drawStretched(0, (v.height() - dupy)*FRACUNIT, v.width()*FRACUNIT/64, dupy*FRACUNIT/64, v.cachePatch("~138"), V_NOSCALEPATCH|V_SNAPTOLEFT|V_SNAPTOBOTTOM)
	v.drawStretched(0, (v.height() - dupy)*FRACUNIT, wscale, dupy*FRACUNIT/64, v.cachePatch("~134"), V_NOSCALEPATCH|V_SNAPTOLEFT|V_SNAPTOBOTTOM)
	v.drawScaled(-2*FRACUNIT, (200 - y)*FRACUNIT, FRACUNIT, v.cachePatch("NGRTIMER"), V_SNAPTOLEFT|V_SNAPTOBOTTOM)
end

hud.add(function(v, p, c)
	if modifiedgame
	and not multiplayer
	and isserver
	and leveltime >= 3*TICRATE
	and not didmodifiedprint
		CONS_Printf(consoleplayer, "\3\131Information:\128 This hub currently does not have a saving system. Feel free to beat the levels in any order, at any time.")
		didmodifiedprint = true
	end
	
	local V = OLDC.Data.voteInfo

	if G_PlatformGametype()
		local p1 = players[0]
		local p2 = players[1]
		
		if OLDC.HUD.notif
			local x = 0
			local interval = TICRATE/3
			local stoppedtime = TICRATE
			
			if OLDC.HUD.notif < interval
				x = -320 * FixedDiv(interval - OLDC.HUD.notif, interval)
			elseif OLDC.HUD.notif > interval + stoppedtime
				x = 320 * FixedDiv(OLDC.HUD.notif - (interval + stoppedtime), interval)
			end
			
			OLDC_DrawFontString(v, x + 160*FRACUNIT, 160*FRACUNIT, 3*FRACUNIT/2, 3*FRACUNIT/2, "Vote failed!", "RFNT", false, 4, 0, 0, SKINCOLOR_FLAME, true)
		end
		
		if V and OLDC.HUD.timer and OLDC.Data.hubSpot
			if CV.vote.value == -1 return end
			if splitscreen and p != p1 return end
			
			DrawVoting(v, p)
		elseif gamemap == hub and not titlemapinaction then
			--v.drawString(0, 192, CURRENT_YEAR.." OLDC Round "..CURRENT_ROUND, V_ALLOWLOWERCASE|V_50TRANS|V_SNAPTOBOTTOM|V_SNAPTOLEFT)
			
			local oldctext = v.cachePatch("OLDCBAR")
			local w = oldctext.width + 4
			local h = oldctext.height
			
			for j = 0, FixedDiv(v.width(), w), FRACUNIT
				v.drawScaled(-w*FRACUNIT + j*w + (leveltime % w)*FRACUNIT, 2*FRACUNIT, FRACUNIT, oldctext, V_70TRANS|V_SNAPTOTOP, v.getColormap(TC_RAINBOW, SKINCOLOR_BONE))
				v.drawScaled(-w*FRACUNIT + j*w + (-leveltime % w)*FRACUNIT, (200 - h - 2)*FRACUNIT, FRACUNIT, oldctext, V_70TRANS|V_SNAPTOBOTTOM, v.getColormap(TC_RAINBOW, SKINCOLOR_BONE))
			end
			
			HUB_PortalTitleDrawer(v, p, c)
		end
	end
end)