local O = OLDC
local F = O.Functions
local CV = O.CVars

local hub = O.hub
local firstmap = O.firstmap
local defaultmap = O.defaultmap

local function getPlayerCount(bots, spect, quit, nobvote)
	local playercount = 0
	
	for p in players.iterate
		if p.bot and not bots continue end
		if p.spectator and not spect continue end
		if p.quittime and not quit continue end
		--if bannedVoters[#p] and nobvote continue end
		
		playercount = $ + 1
	end
	
	return playercount
end

local function levelHUDdisables(m)
	if mapheaderinfo[m].hidetitle
		F.DEShortcut("stagetitle", false)
	else
		F.DEShortcut("stagetitle", true)
	end

	if mapheaderinfo[m].creditsmap or mapheaderinfo[m].marathonstartup or m == hub
		F.DEShortcut("score", false)
		F.DEShortcut("time", false)
		F.DEShortcut("rings", false)
		F.DEShortcut("lives", false)
	else
		F.DEShortcut("score", true)
		F.DEShortcut("time", true)
		F.DEShortcut("rings", true)
		F.DEShortcut("lives", true)
	end
end

local function failVote()
	local D = O.Data
	
	if D.voteInfo.player and D.voteInfo.player.valid
		if multiplayer -- KIMOKAWAIII has level portals instead, so you don't get caught in the voting sector
			if D.hubSpot and D.hubSpot.valid
				P_TeleportMove(D.voteInfo.player.mo, D.hubSpot.x, D.hubSpot.y, D.hubSpot.z)
				D.voteInfo.player.mo.angle = D.hubSpot.angle
			else
				G_DoReborn(#D.voteInfo.player)
			end

			S_StartSound(D.voteInfo.player.mo, sfx_mixup, D.voteInfo.player)
			--P_FlashPal(O.voteInfo.player, 1, 5)
		end
	end
end

local function finishVote(naturalend)
	local D = O.Data
	
	if not D.voteInfo.voting return end
	
	D.voteInfo.endtimer = $ and $ + 1 or 1
	
	if (naturalend and G_PlatformGametype())
		if D.voteInfo.endtimer < TICRATE/2 return end
		
		if F.getVoteCount("yes") > F.getVoteCount("no") or CV.vote.value == -1
			F.ExitLevel(D.voteInfo.map, 1)
		else
			D.voteInfo.endtimer = 0
			
			if multiplayer
				O.HUD.notif = 1
			end
			
			failVote()
		end
	end
	
	D.voteInfo.voting = false
	//D.voteInfo.map = defaultmap
	
	levelHUDdisables(gamemap)
end

local function startVote(mo, map)
	if not G_PlatformGametype() return end
	
	local D = O.Data
	
	if D.voteInfo.voting return end
	
	if not mo return end
	if not mo.valid return end
	if not mo.player return end
	
	local p = mo.player
	
	if p.bot or p.spectator return end
	--if (bannedVoters[#player]) return end
	
	if CV.pick.value
		if not (p == server or player == admin)
			CONS_Printf(p, "The host has disallowed others from picking levels.")
			return
		end
	end
	
	D.voteInfo.player = p
	
	if (CV.vote.value > 0)
		D.voteInfo.timer = CV.vote.value * TICRATE
	elseif (CV.vote.value == 0)
		D.voteInfo.timer = 12
	end
	
	if (map > 0)
		D.voteInfo.map = map
	else
		D.voteInfo.map = defaultmap
	end
	
	O.HUD.scroll = 320*FRACUNIT
	O.HUD.inittimer = leveltime
	O.HUD.notif = 0
	
	D.playerVotes = {}
	D.voteInfo.voting = true

	F.DEShortcut("score", false)
	F.DEShortcut("time", false)
	F.DEShortcut("rings", false)
	F.DEShortcut("lives", false)
	
	//S_StartSound(toucher, sfx_3db16)
	S_StartSound(toucher, sfx_gravch)
end

local function loadClearedMaps()
	local D = O.Data
	
	for i = 0, TOTALENTRIES - 1
		-- To do: fix this for the next hub
		-- It's supposed to bitshift 1, not 2
		if (OLDCsave[1] & (1 << i))
			D.mapChecks[i + firstmap] = true
		end
	end
end

local function levelPortalSpawn(mo)
	local D = O.Data
	
	local map = AngleFixed(mo.angle) / FRACUNIT

	if not (mapheaderinfo[map + firstmap - 1])
		P_RemoveMobj(mo)
		return
	end

	if (map > 0)
		if (map > 52)
			mo.sprite = SPR_LVLC
		elseif (map > 26)
			mo.sprite = SPR_LVLB
		end
		mo.frame = $ | ((map-1) % 26)
	end

	if not (mapheaderinfo[map + firstmap - 1].typeoflevel & F.GetTOL(gametype))
		mo.portaldisable = 1
		mo.flags2 = $1|MF2_SHADOW
	end

	P_SpawnMobj(mo.x, mo.y, mo.z, MT_LEVELLIGHT)
	mo.portalz = mo.z
	mo.z = $ + 64*FRACUNIT
	mo.spawned = true
end

addHook("MobjThinker", function(mo)
	if not (mo.spawned) -- because mobjspawn hook caused unknown weirdness
		if (leveltime > 2) -- wait for level load linedefs
			levelPortalSpawn(mo)
		end
		return
	end

	if (mo.portalz == nil)
		mo.portalz = mo.floorz
	end

	if (mo.lvlportal == nil)
		mo.lvlportal = {}
	end

	local lvl = (AngleFixed(mo.angle) / FRACUNIT) + firstmap - 1

	local amp = 8
	local speed = 2*TICRATE
	local pi = 22*FRACUNIT/7
	local sine = amp * sin((2*pi*speed) * (leveltime + (lvl^2)))
	mo.z = (mo.portalz+64*FRACUNIT) + sine
	mo.spriteyoffset = 32*FRACUNIT

	local bossoffset = 0
	if (mo.portaldisable)
		bossoffset = 6
	elseif (mapheaderinfo[lvl] and mapheaderinfo[lvl].Fboss)
		bossoffset = 3
	end

	if not (mo.lvlportal[1] and mo.lvlportal[1].valid)
		mo.lvlportal[1] = P_SpawnMobj(mo.x, mo.y, mo.portalz, MT_LEVELPORTAL)
		mo.lvlportal[1].target = mo
	end
	mo.lvlportal[1].state = S_LEVELPORTAL1 + ((leveltime/3) % 3) + bossoffset

	local numlayers = 2 -- 1
	local curlayer = 0
	local numparticles = 3
	local curparticle = 2
	local speed = 20
	local rad = 24

	while (curlayer < numlayers)
		for i = curparticle,curparticle+numparticles
			if not (mo.lvlportal[i] and mo.lvlportal[i].valid)
				mo.lvlportal[i] = P_SpawnMobj(mo.x, mo.y, mo.portalz, MT_LEVELPORTAL)
				mo.lvlportal[i].angle = FixedAngle((360*FRACUNIT/numparticles) * (i-(curparticle-1))) - (ANG15*curlayer)
			else
				A_Custom3DRotate(mo.lvlportal[i], rad, speed<<16)
			end
			mo.lvlportal[i].state = S_LEVELPORTAL1 + (((leveltime/3)+curlayer+1) % 3) + bossoffset
			mo.lvlportal[i].target = mo
			mo.lvlportal[i].z = mo.portalz
		end

		curparticle = curparticle+numparticles+1
		--numparticles = $+2
		--speed = 3*$/2
		rad = $+24
		curlayer = $+1
	end

	if (mo.reactiontime > 0)
		mo.reactiontime = $1-1
	end

	if (leveltime % 3 == 0)
	and not (O.Data.mapChecks[lvl])
	and not (mo.portaldisable)
		local hdist = P_RandomRange(0,80) * FRACUNIT
		local hang = FixedAngle(P_RandomRange(0,359) * FRACUNIT)
		local vdist = P_RandomRange(0,64) * FRACUNIT
		local vang = FixedAngle(P_RandomRange(-180,180) * FRACUNIT)

		local sparkle = P_SpawnMobj(
			mo.x + FixedMul(FixedMul(hdist, cos(hang)), cos(vang)),
			mo.y + FixedMul(FixedMul(hdist, sin(hang)), cos(vang)),
			mo.z + 32*FRACUNIT + FixedMul(vdist, sin(vang)),
			MT_LEVELSPARKLE
		)

		sparkle.scale = 2*FRACUNIT
		sparkle.destscale = sparkle.scale

		if (mapheaderinfo[lvl] and mapheaderinfo[lvl].Fboss)
			sparkle.state = S_BOSSSPARKLE1
		end
	end
end, MT_LEVELSPOT)

addHook("TouchSpecial", function(mo, toucher)
	if mo.portaldisable
		if toucher.player and not mo.reactiontime
			if mo.portaldisable == 1
				local gtstr = "Cooperative" -- too lazy to add the others
				CONS_Printf(toucher.player, "This level does not support "..gtstr.." mode.")
			elseif mo.portaldisable == 2
				CONS_Printf(toucher.player, "This level needs to be unlocked.")
			end
		end
		
		mo.reactiontime = TICRATE
	else
		local lvl = FixedInt(AngleFixed(mo.angle)) + firstmap - 1
		
		startVote(toucher, lvl)
	end
	
	if (leveltime & 1) and toucher.player.speed
		local ring = P_SpawnMobjFromMobj(toucher, 0, 0, toucher.height/2, MT_THOK)
		ring.tics = -1
		ring.sprite = SPR_CIRC
		ring.frame = FF_PAPERSPRITE
		ring.angle = toucher.angle + ANGLE_90
		ring.color = SKINCOLOR_TEAL
		ring.fuse = TICRATE/2
		ring.momz = 4*FRACUNIT
		ring.scale = mo.scale/32
		ring.destscale = 2*FRACUNIT/3
		ring.scalespeed = (ring.destscale - ring.scale)/ring.fuse
	end
	
	return true
end, MT_LEVELSPOT)

addHook("MobjThinker", function(mo)
	if not G_PlatformGametype() return end
	
	local D = O.Data
	local HUD = O.HUD
	
	D.hubSpot = $ or mo
	
	HUD.notif = $ and min(2*TICRATE, $ + 1) or 0
	
	if D.voteInfo.voting
		HUD.scroll = max(FRACUNIT, $ - ($ / 8))
		HUD.timer = min(TICRATE/3, $ + 1)
	elseif not D.voteInfo.endtimer
		HUD.scroll = $ > 0 and $ * -1 or $
		
		HUD.scroll = max(-320*FRACUNIT, $ - abs($ * 2))
		HUD.timer = max(0, $ - 1)
		return
	end
	
	if (D.voteInfo.timer > 0 and CV.vote.value > 0)
	or CV.vote.value == 0
		for p in players.iterate
			p.pflags = $ | PF_FULLSTASIS
				
			local buffertime = 2
			
			if CV.vote.value > 0
				buffertime = (CV.vote.value * TICRATE) - 10
			end
				
			if (D.voteInfo.timer >= buffertime and p.cmd.buttons) or D.voteInfo.endtimer
				p.disablevote = true
			elseif (p.disablevote) and not (p.cmd.buttons)
				p.disablevote = false
			end
			
			local oldvote = D.playerVotes[#p]
			
			--if (O.playerVotes[#player] == nil) and not (player.disablevote)
			if not p.disablevote
			--and not (player.bot or player.spectating or bannedVoters[#player])
			and not (p.bot or p.spectating)
				if (p.cmd.buttons & BT_JUMP) and not (p.cmd.buttons & BT_USE)
					D.playerVotes[#p] = true
				elseif (p.cmd.buttons & BT_USE) and not (p.cmd.buttons & BT_JUMP)
					D.playerVotes[#p] = false
				end
			end
			
			if oldvote != D.playerVotes[#p]
				p.voteblink = TICRATE
				S_StartSound(nil, sfx_kc48, p)
			end
		end
	end
	
	if D.voteInfo.timer <= 0 or F.getVoteCount() >= getPlayerCount(false, false, false)
		finishVote(true)
	elseif (CV.vote.value > 0 and D.voteInfo.timer > 0) or (CV.vote.value == 0 and D.voteInfo.timer > 1)
		D.voteInfo.timer = $ - 1
	end
end, MT_HUBSPOT)

addHook("MapLoad", function(map)
	local D = O.Data
	
	for thing in mapthings.iterate
		if not (thing.mobj and thing.mobj.valid) continue end
		
		if thing.mobj.type == MT_HUBSPOT
			D.hubSpot = thing.mobj
		end
	end
	
	if D.hubSpot
		if not (multiplayer or modifiedgame)
			for i = 1, 3
				P_LinedefExecute(3000+i)
			end
		end
		
		D.lasthub = map
	end

	for p in players.iterate	
		p.awayviewmobj = nil -- vanilla doesn't reset these already? wut :V
		p.awayviewtics = 0
		p.awayviewaiming = 0
	end
	
	levelHUDdisables(map)
	
	if (map == hub and marathonmode) then
		resetLives()
		if (marathonmode) then
			F.ExitLevel(O.firstmap, 1)
			return
		end
	end
	
	G_SetCustomExitVars(F.NextLevel(map))
end)

addHook("MapChange", function(map)
	local D = OLDC.Data
	
	D.hubSpot = nil
	D.failedlevel = false
	D.exitingcheck = false
	
	D.voteInfo.endtimer = 0
	
	OLDC.HUD.scroll = 320*FRACUNIT
	OLDC.HUD.timer = 0
	OLDC.HUD.notif = 0
	
	finishVote(false)

	levelHUDdisables(map)
	
	if map == hub and not multiplayer
		loadClearedMaps()
	end
end)

addHook("ThinkFrame", do
	if gamestate == GS_TITLESCREEN
		OLDC.Data.mapChecks = {}
	end
end)

addHook("PlayerSpawn", function(p)
	if (gamemap ~= hub) then return end
	
	p.pflags = $ & ~(PF_FORCESTRAFE|PF_NOCLIP)
	
	local good, valid = pcall(do return p.realmo and p.realmo.valid end)
	if (good and valid) then
		p.jumpfactor = skins[p.mo.skin].jumpfactor
		p.charability = skins[p.mo.skin].ability
		p.charability2 = skins[p.mo.skin].ability2
		p.normalspeed = skins[p.mo.skin].normalspeed
	end	
end)

addHook("PlayerThink", function(p)
	if CV.nolives.value then
		if p.bot return end
		
		p.lives = INFLIVES
	end
	
	p.voteblink = $ and $ - 1 or 0
end)

addHook("ThinkFrame", do
	local D = O.Data
	local allexiting = true
	
	for p in players.iterate
		if p.bot continue end
		
		if not p.exiting
			allexiting = false
		end
	end
	
	D.exitingcheck = allexiting
	
	if D.exitingcheck
	and not D.failedlevel
	and gamemap >= firstmap
	and gamemap <= firstmap + TOTALENTRIES
	and not D.mapChecks[gamemap]
		D.mapChecks[gamemap] = true
		
		if not multiplayer
			-- To do: fix this for the next hub
			-- It's supposed to bitshift 1, not 2
			OLDCsave[1] = $ | (1 << (gamemap - firstmap))
		end
	end
	
	D.pausetimer = false
end)

--addHook("MobjSpawn", levelPortalSpawn, MT_LEVELSPOT)

local hudview = 0
local spinpress = 0
local tabpress = false

local function ToggleTabHUD(toggle)
	local func = toggle and hud.enable or hud.disable
	
	func("rankings")
	func("coopemeralds")
	func("tokens")
	func("tabemblems")
end

addHook("PlayerThink", function(p)
	if p.bot return end
	
	if modeattacking
	or marathonmode
	or not hub
	or gamemap == hub
	or multiplayer
		ToggleTabHUD(true)
		return 
	end
	
	if (p.cmd.buttons & BT_SPIN)
	and tabpress
	and not p.playerstate
		hudview = min($ + 1, 5)
		spinpress = min($ + 1, 3*TICRATE + 1)
		
		ToggleTabHUD(false)
	else
		hudview = max(0, $ - 1)
		
		ToggleTabHUD(true)
	end
	
	if not hudview
		spinpress = 0
	end
	
	if spinpress == 3*TICRATE
		spinpress = $ + 3*TICRATE + 1
		COM_BufInsertText(p, "abortlevel yes")
	end
end)

addHook("MapChange", do
	hudview = 0
	spinpress = 0
	tabpress = false
end)

local function DrawQuickReturnPrompt(v, x, y, s)
	local trans = 0
	local yoffs = 11*s
	y = $ - 2*yoffs - 8*s
	
	OLDC_DrawFontString(v, x, y, s, s, "Hold", "RFNT", false, 1, trans, 0, SKINCOLOR_WHITE, true)
	y = $ + yoffs
	OLDC_DrawFontString(v, x, y, s*2, s*2, "Tab + Spin", "RFNT", false, 1, trans, 0, SKINCOLOR_YELLOW, true)
	y = $ + yoffs*2
	OLDC_DrawFontString(v, x, y, s, s, "to return", "RFNT", false, 1, trans, 0, SKINCOLOR_WHITE, true)
	y = $ + yoffs
	OLDC_DrawFontString(v, x, y, s, s, "to the hub", "RFNT", false, 1, trans, 0, SKINCOLOR_WHITE, true)
end

hud.add(function(v, p)
	tabpress = false
	
	if not hudview return end
	
	v.fadeScreen(31, hudview)
	
	local x = 160*FRACUNIT
	local y = 72*FRACUNIT
	local trans = (5 - hudview) * V_10TRANS
	local circ = 360*FixedDiv(min(spinpress, 3*TICRATE), 3*TICRATE + 1)
	local s = abs(sin(FixedAngle(spinpress*FRACUNIT*4)))/8 + 7*FRACUNIT/8
	
	for i = 0, (FixedInt(circ) % 360), 5
		v.drawScaled(
			x,
			y - (i == 0 and 2*s or 0),
			s/2,
			v.getSpritePatch(SPR_CIRC, B, 0, (360 - i)*ANG1),
			trans,
			v.getColormap(p.skin, p.skincolor)
		)
	end
	
	DrawQuickReturnPrompt(v, x, y, s)
end, "game")

hud.add(function(v)
	if modeattacking or marathonmode return end
	if not hub return end
	if gamemap == hub return end
	if gamemap < firstmap or gamemap > firstmap + TOTALENTRIES -1 return end
	if multiplayer return end
	
	tabpress = true
	
	if hudview return end
	
	local x = 160*FRACUNIT
	local y = 72*FRACUNIT
	local s = 7*FRACUNIT/8
	
	DrawQuickReturnPrompt(v, x, y, s)
end, "scores")